#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#ifndef SIGNAL
#include <avr/signal.h>
#endif // SIGNAL
#include <avr/iom128.h>
#include "inc/constants.h"
#include "inc/comm.h"
#include "inc/uart.h"

// Mgliche Antworten: ACK/NACK/PR_MISMATCH
unsigned char ACK           = 0x0F;
unsigned char NACK          = 0xF0;
unsigned char PR_MISMATCH   = 0xAA;

/**
* WaitACK: Wenn TRUE, dann das aktuelle Device wartet auf einem ACK von der Gegestelle
*/
volatile int WaitACK = FALSE;

/**
* data: Die empfangenen Daten von der Gegenstelle
*/
struct dataFromRS232 data;

/**
* dataQueue: Die zu versendenen Datenqueue
*/
struct dataFromRS232 dataQueue[QUEUESIZE];

/**
* commInitialized: TRUE wenn die Initialisierung erfolgreich war, FALSE ansonsten
*/
boolean commInitialized = FALSE;

/**
 * Anzahl der Fehler bei der Kommunikation ber RS232
 */
uint8_t rs232ErrorCount;

#ifdef USE_UART0
  #define u_init    uart_init
  #define u_getc    uart_getc
  #define u_putc    uart_putc
  #define u_puts    uart_puts
  #define u_puts_p  uart_puts_p
  #define u_puts_P  uart_puts_P
#endif
#ifdef USE_UART1
  #define u_init    uart1_init
  #define u_getc    uart1_getc
  #define u_putc    uart1_putc
  #define u_puts    uart1_puts
  #define u_puts_p  uart1_puts_p
  #define u_puts_P  uart1_puts_P
#endif

/**
 * Schickt ein ACK
 */
#define sendACK()         u_putc(ACK)

/**
 * Schickt ein NACK
 */
#define sendNACK()        u_putc(NACK)

/**
 * Schickt ein PR_MISMATCH
 */
#define sendPRMismatch()  u_putc(PR_MISMATCH)

/**
 * Beginnt die Wartezeit auf ACK
 */
#define setWaitACK()      { WaitACK = TRUE; }

/**
 * Beendet die Wartezeit auf ACK
 */
#define clearWaitACK()    { WaitACK = FALSE; }

/**
 * TRUE wenn das Programm auf ACK wartet
 */
#define isWaitingACK()    (WaitACK == TRUE)

/**
 * Berechnet die Paritt fr das empfangenen Packet
 *
 * @param struct dataFromRS232 data  Die empfangenen Daten
 * @return uint8_t                   Die Paritt
 */
uint8_t calculateParity(struct dataFromRS232 data)
{
  uint8_t c, parity;

  parity = data.deviceFrom + data.deviceTo + data.packetSize;
  for(c = 0; c < data.packetSize; c++)
    parity += data.data[c];
  parity = parity % 256;

  return(parity);
}

void sendData(struct dataFromRS232 data)
{
  data.parity = calculateParity(data);
  u_putc(MYPROTOCOLVERSION);
  u_putc(data.deviceFrom);
  u_putc(data.deviceTo);
  u_putc(data.packetSize);
  u_puts(data.data);
  u_putc(data.parity);
  setWaitACK();
}

void doInitCommunication()
{
  u_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
  clearWaitACK();
  rs232ErrorCount = 0;
  ackWaitCount = 0;

  data.version = MYPROTOCOLVERSION;
  data.deviceFrom = 3;
  data.deviceTo = 1;
  strcpy(data.data, "READY");
  data.packetSize = 5;
  sendData(data);
}

void doLoopJobsComm()
{
  unsigned int c;
  uint8_t l;

  c = u_getc();
  if((c & 0xFF00) == 0)
  {
    if(isWaitingACK())
    {
      clearWaitACK();
    }
    else
    {
      receivedData.version = (uint8_t) (c & 0xFF);
      do { c = u_getc(); } while(c & UART_NO_DATA);
      receivedData.deviceFrom = (uint8_t) (c & 0xFF);
      do { c = u_getc(); } while(c & UART_NO_DATA);
      receivedData.deviceTo = (uint8_t) (c & 0xFF);
      do { c = u_getc(); } while(c & UART_NO_DATA);
      receivedData.packetSize = (uint8_t) (c & 0xFF);
      for(l = 0; l < receivedData.packetSize; l++)
      {
        do { c = u_getc(); } while(c & UART_NO_DATA);
        receivedData.data[l] = (char) (c & 0xFF);
      }
      receivedData.data[l] = 0;
      do { c = u_getc(); } while(c & UART_NO_DATA);
      receivedData.parity = (uint8_t) (c & 0xFF);
      if(receivedData.version != MYPROTOCOLVERSION)
        sendNACK();
      else if(calculateParity(receivedData) != receivedData.parity)
        sendPRMismatch();
      else
      {
        sendACK();

        data.version = MYPROTOCOLVERSION;
        data.deviceFrom = receivedData.deviceTo;
        data.deviceTo = receivedData.deviceFrom;
        strcpy(data.data, receivedData.data);
        data.packetSize = strlen(data.data);
        sendData(data);
      }
    }
  }
} 
