#include <math.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#ifndef SIGNAL
#include <avr/signal.h>
#endif // SIGNAL
#include "inc/constants.h"
#include "inc/i2cmaster.h"
#include "inc/tools.h"

/**
* So viele IRQs werden jede Sekunde ausgelst.
* Fr optimale Genauigkeit muss
* IRQS_PER_SECOND ein Teiler von F_CPU sein
* und IRQS_PER_SECOND ein Vielfaches von 100.
* Ausserdem muss gelten F_CPU / IRQS_PER_SECOND <= 65536
*/
#define IRQS_PER_SECOND   1000L   /* 10 s */

/**
* Anzahl IRQs pro 10 Millisekunden
*/
#define IRQS_PER_10MS     (IRQS_PER_SECOND / 100)

/**
* Gltigkeitsprfung.
* Bei ungeeigneten Werten gibt es einen Compilerfehler
*/
#if (F_CPU/IRQS_PER_SECOND > 65536) || (IRQS_PER_10MS < 1) || (IRQS_PER_10MS > 255)
#   error Diese Werte fuer F_CPU und IRQS_PER_SECOND
#   error sind ausserhalb des gueltigen Bereichs!
#endif

/**
* Compiler-Warnung falls die Genauigkeit nicht optimal ist.
* Wenn das nervt fr deine Werte, einfach lschen :-)
*/
#if (F_CPU % IRQS_PER_SECOND != 0) || (IRQS_PER_SECOND % 100 != 0)
#   warning Das Programm arbeitet nicht mit optimaler Genauigkeit.
#endif

#if !defined (TCNT1H)
#error Dieser Controller hat keinen 16-Bit Timer1!
#endif // TCNT1H


/*************************************************************************
 delay loop for small accurate delays: 16-bit counter, 4 cycles/loop
*************************************************************************/
inline void _delayFourCycles(unsigned int __count)
{
  if ( __count == 0 )
    __asm__ __volatile__( "rjmp 1f\n 1:" );       // 2 cycles
  else
    __asm__ __volatile__ (
                          "1: sbiw %0,1" "\n\t"
                          "brne 1b"               // 4 cycles/loop
                          : "=w" (__count)
                          : "0" (__count)
    );
}

/**
* Die Interrupt Service Routine (ISR).
* Die Variabel "uptime" wird erhht
*/
SIGNAL (SIG_OUTPUT_COMPARE1A)
{
  uptime++;
}

/**
* Initialisiert den Timer
*/
void doInitTimer()
{
  uptime = 0L;

  // Timer1: keine PWM
  TCCR1A = 0;

  // Timer1 ist Zhler: Clear Timer on Compare Match (CTC, Mode #4)
  // Timer1 luft mit vollem MCU-Takt: Prescale = 1
#if defined (CTC1) && !defined (WGM12)
  TCCR1B = (1 << CTC1)  | (1 << CS10);
#elif !defined (CTC1) && defined (WGM12)
  TCCR1B = (1 << WGM12) | (1 << CS10);
#else
#error Keine Ahnung, wie Timer1 fuer diesen AVR zu initialisieren ist!
#endif

  // OutputCompare fr gewnschte Timer1 Frequenz
  // TCNT1 zhlt immer 0...OCR1A, 0...OCR1A, ... 
  // Beim berlauf OCR1A -> OCR1A+1 wird TCNT1=0 gesetzt und im nchsten
  // MCU-Takt eine IRQ erzeugt.
  OCR1A = (unsigned short) ((unsigned long) F_CPU / IRQS_PER_SECOND-1);

  // OutputCompareA-Interrupt fr Timer1 aktivieren
#if defined (TIMSK1)
  TIMSK1 |= (1 << OCIE1A);
#elif defined (TIMSK)
  TIMSK  |= (1 << OCIE1A);
#else
#error Keine Ahnung, wie IRQs fuer diesen AVR zu initialisieren sind!
#endif
}

SIGNAL(SIG_OUTPUT_COMPARE3A)       
{
  usekStart++;
}

// Interrupt alle 10s
void initTimer3()
{
  usekStart = 0L;
  degrees = 0L;

  OCR3A  = 80;
  OCR3B  = 800;
  TCCR3B = (1<<WGM32) | (1<<CS30);
  ETIMSK = (1<<OCIE3A) | (1<<OCIE3B);
}

/**
* Warte x Millisekunden
*
* @param uint8_t x   Die Millisekunden zu warten
*/
void delayMS(unsigned int ms)
{
  int l;

  for(l = 0; l < ms; l++)
    delayUS(1000);
}

/**
* Rundet die angegebene Zahl
*
* @param uint8_t x   Die zu rundenden Zahl
*/
double round(double x)
{
  double intPart;
  double decPart;

  intPart = floor(x);
  decPart = x - intPart;
  if(decPart > 0.5)
    return (intPart + 1);
  else
    return intPart;
}

/**
* Sendet ein Byte an die angegebenen I2C-Device
*
* @param  uint8_t addr  Die I2C-Device
* @param  uint8_t i     Die Information (Byte) zu schicken
* @return boolean       TRUE wenn alles in Ordnung, FALSE falls Probleme
*/
boolean writeI2C(uint8_t addr, uint8_t i)
{
  if(i2c_start(addr + I2C_WRITE) != 0)
    return FALSE;
  if(i2c_write(i) != 0)
    return FALSE;
  i2c_stop();

  return TRUE;
}
